open Expr_lexer
open Expr_parser
open Expr_conv
open Sharing
open Graph

let display_solution = false

let examples : string list = 
  ["x = x"; 
   "f(x,y) = f(x,x)"; 
   "succ(x) = succ(succ(zero()))";
   "succ(x) = succ(succ(y))";
   "g(x) = f(x,y)";
   "f(x,g(y,y),x) = f(z,z,g(w,f(t)))"; 
   "f(x,g(z,z),z) = f(g(y,y),y,g(w,w))";
   "succ(plus(x,succ(y))) = succ(plus(succ(t),z))";
   "x = y
    y = x";
   "z = t
    y = z
    x = y";
   "x = y
    y = z
    z = t";
]

let big n = 
  let rec rhs acc n =
    if n = 0 then acc
    else
      rhs ((Printf.sprintf "g(x%d, x%d)" n n):: acc) (n - 1)
  in 
  let rec lhs acc n = 
    if n = 0 then acc else
      lhs ((Printf.sprintf "x%d" (n - 1))::acc) (n - 1)
  in 
  Printf.sprintf "f(%s) = f(%s)" 
  (String.concat ", " (lhs [] n)) (String.concat ", " (rhs [] n))

let examples : string list = 
  examples@[big 10]

(* Les exemples qui feront probablement diverger vos solutions: *)
let examples : string list = examples@["x = f(x)"; "x = f(y) y=f(x)"]

let examples : Expr.problem list = 
  List.map Expr_conv.problem_of_string examples

let examples : Sharing.problem list = List.map 
  (fun problem -> 
    List.fold_left (fun (problem_sharing, defined) (x, y) ->
    let x, defined = Sharing.sharing_of_expr defined x in 
    let y, defined = Sharing.sharing_of_expr defined y in 
    ((x,y)::problem_sharing,defined)) ([], []) problem)
    examples

let solve ((eqns, defined) as problem : Sharing.problem) = 
  Printf.printf "Solving : \n%s\n" (string_of_problem problem);
  try
    flush stdout;
    List.iter (fun (x,y) -> unify x y) eqns;
    Printf.printf "Solution : \n";
    List.iter (fun (var, r) -> 
      Printf.printf "%s -> %s %s\n" var 
      (string_of_sharing_swapped defined r) 
      (if cyclic r then "(cyclic !)" else "")) defined; 
    if display_solution then
      ignore (Graph.display_solution defined)
  with Not_unifyable -> Printf.printf "Not unifyable.\n"

let _ = 
  List.iter solve examples

    (*
let _ = 
  List.iter Graph.display
    (List.map fst (List.map Sharing.sharing_of_expr (List.map snd (List.flatten
    examples)))) *)



