type symbol = string
type var = string 

type descr =
| Link of  t
| Repr of constr

and constr =
| Var
| Op of symbol  * t list

and t = descr ref

type problem = (t * t) list * (var * t) list

let rec sharing_of_expr defined = function
  | Expr.Var x -> (try 
        (List.assoc x defined, defined)
      with Not_found -> 
        let r = ref (Repr Var) in 
        r, (x, r)::defined)
  | Expr.Op (f, l) -> 
      let (l, defined) = 
        List.fold_left (fun (l, defined) x -> 
          let r, defined = 
            sharing_of_expr defined x in (r::l, defined)) ([],defined) l 
          in (ref (Repr (Op (f, l))), defined)

let rec repr t = match !t with 
  | Link t' -> let r = repr t' in t := Link r; r
  | Repr r -> t

exception Not_unifyable 

let rec unify t1 t2 = 
  let r1, r2 = repr t1, repr t2 in 
  match !r1, !r2 with 
  | Link _, _ | _, Link _ -> assert false
  | Repr Var, Repr Var when r1 == r2-> ()
  | Repr Var, _ -> r1 := Link r2
  | _, Repr Var -> r2 := Link r1
  | Repr Op (f1, args1), Repr Op (f2, args2) when f1 = f2 -> 
        (try
          List.iter2 unify args1 args2
        with Invalid_argument "List.iter2" -> raise Not_unifyable)
  | _ -> raise Not_unifyable 

let expr_of_sharing names r =
  let rec aux r =
    match !r with 
    | Link r -> aux r
    | Repr Var -> 
        Expr.Var (List.assq r names)
    | Repr (Op (f, t)) -> 
        Expr.Op (f, List.map aux t)
  in aux r

let string_of_sharing defined t = 
 let rec aux t = 
    try 
      List.assq t defined 
    with Not_found -> begin
      match !t with 
       | Link r -> aux r
       | Repr (Op (f, args)) -> Printf.sprintf "%s(%s)" f
           (String.concat ", " (List.map aux args))
       | Repr Var -> failwith "string_of_sharing"
    end
  in begin
      match !t with 
       | Link r -> aux r
       | Repr (Op (f, args)) -> Printf.sprintf "%s(%s)" f
           (String.concat ", " (List.map aux args))
       | Repr Var -> List.assq t defined 
    end

let cyclic t = 
  let rec aux visited t = 
    if List.memq t visited then true
    else match !(repr t) with 
      | Repr (Op (f, args)) -> 
          List.exists (aux (t::visited)) args
      | _ -> false
  in aux [] t

let string_of_sharing_swapped defined t = 
  string_of_sharing (List.map (fun (x,y) -> (y,x)) defined) t

let string_of_problem (eqns, defined) = 
    String.concat "\n" 
      (List.map (fun (x,y) -> 
        Printf.sprintf "%s = %s" 
         (string_of_sharing_swapped defined x)
         (string_of_sharing_swapped defined y)) eqns)
